#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright © 2024 Ye Chang yech1990@gmail.com
# Distributed under terms of the GNU license.
#
# Created: 2024-06-08 20:32


import os
import random

from . import mappy as mp
from .conversion import km_conversion, mk_conversion, mk_convert_file


def find_properly_paired_hits(hits, fwd=True):
    """
    # find properly paired hits
    Given a list of hits with read_num, ref_name, ref_start, ref_end, strand. For example:
        1 rRNA-Hsa-nucleus_locus 8204 8237 1
        1 rRNA-Hsa-nucleus_locus 21555 21588 1
        2 rRNA-Hsa-nucleus_locus 21561 21596 -1
        2 rRNA-Hsa-nucleus_locus 8210 8245 -1

    This function will find all properly paired hits. A hit is properly paired if it is properly paired.
    We defined properly paired hits as:
    1. The hit paris are from two reads (read_num=1 and read_num=2)
    2. Both hits mapped to the same reference sequence
    3. read 1 and reads 2 of one pair are in different strand (if reads 1 is 1, reads 2 is -1)
    4. If fwd is True, strand = 1 is forward, strand = -1 is reverse. Otherwise, strand = 1 is reverse, strand = -1 is forward
    5. The mapping start of read 1 is smaller than the mapping end of read 2, if fwd is True. Otherwise, the mapping end of read 1 is larger than the mapping start of read 2.
    6. The distance between the mapping start and end of read 1 and read 2 is less than 1000
    """
    parsed_hits = []
    # group by ref_name and separate read 1 and read 2
    ref_name_hits = {}
    for hit in hits:
        if hit.ctg not in ref_name_hits:
            ref_name_hits[hit.ctg] = [[], []]
        ref_name_hits[hit.ctg][hit.read_num - 1].append(hit)
    for hits in ref_name_hits.values():
        if len(hits[0]) > 0 and len(hits[1]) > 0:
            for hit1 in hits[0]:
                for hit2 in hits[1]:
                    if hit1.strand + hit2.strand == 0:
                        if fwd:
                            if hit1.r_st < hit2.r_en and hit2.r_en - hit1.r_st < 1000:
                                parsed_hits.append((hit1, hit2))
                        else:
                            if hit1.r_en > hit2.r_st and hit1.r_en - hit2.r_st < 1000:
                                parsed_hits.append((hit1, hit2))

    return parsed_hits


def cal_md_and_tag(cigar, seq, ref, fwd):
    """
    fwd:
    if True, A->G, C->T
    if False, T->C, G->A

    calculate MD tag
    M (0): Alignment match
    I (1): Insertion to the reference
    D (2): Deletion from the reference
    N (3): Skipped region from the reference
    S (4): Soft clipping
    H (5): Hard clipping
    P (6): Padding
    = (7): Sequence match
    X (8): Sequence mismatch

    Some tags:
    Yf:i:<N>: Number of (A to G) conversions are detected in the read.
    Zf:i:<N>: Number of (A to A) un-converted bases are detected in the read. Yf + Zf = total number of bases which can be converted in the read sequence.
    Yc:i:<N>: Number of (C to T) conversions are detected in the read.
    Zc:i:<N>: Number of (C to C) un-converted bases are detected in the read. Yc + Zc = total number of bases which can be converted in the read
    NS:i:<N>: Number of substitutions are detected in the read.
    NC:i:<N>: Number of clipped bases and INDEL bases are detected in the read.


    ### add tag function that I used previously, learn from this function
    for read_pos, _, ref_base in read.get_aligned_pairs(with_seq=True):
        # Dn not forget to convert the ref_base to upper case
        ref_base = ref_base.upper() if ref_base is not None else None
        read_base = s[read_pos] if read_pos is not None else None
        if strand == "+":
            if ref_base == "A":
                if read_base == "G":
                    yf += 1
                elif read_base == "A":
                    zf += 1
                elif read_base is not None:
                    ns += 1
            elif ref_base == "C":
                if read_base == "T":
                    yc += 1
                elif read_base == "C":
                    zc += 1
                elif read_base is not None:
                    ns += 1
            elif ref_base is None or read_base is None:
                nc += 1
            elif ref_base != read_base:
                ns += 1

        else:
            if ref_base == "T":
                if read_base == "C":
                    yf += 1
                elif read_base == "T":
                    zf += 1
                elif read_base is not None:
                    ns += 1
            elif ref_base == "G":
                if read_base == "A":
                    yc += 1
                elif read_base == "G":
                    zc += 1
                elif read_base is not None:
                    ns += 1
            elif ref_base is None or read_base is None:
                nc += 1
            elif ref_base != read_base:
                ns += 1
    """
    yf = 0
    zf = 0
    yc = 0
    zc = 0
    ns = 0
    nc = 0
    md_tag = []
    ref_index = 0
    query_index = 0
    match_count = 0

    if fwd:
        b1, b2, b3, b4 = "A", "G", "C", "T"
    else:
        b1, b2, b3, b4 = "T", "C", "G", "A"

    for length, operation in cigar:
        if operation == 0:  # Match or Mismatch
            for i in range(length):
                if ref[ref_index] == seq[query_index]:
                    match_count += 1
                    if seq[query_index] == b1:
                        zf += 1
                    elif seq[query_index] == b3:
                        zc += 1
                else:
                    if match_count > 0:
                        md_tag.append(str(match_count))
                        match_count = 0
                    md_tag.append(ref[ref_index])
                    if seq[query_index] == b2:
                        yf += 1
                    elif seq[query_index] == b4:
                        yc += 1
                    else:
                        ns += 1
                ref_index += 1
                query_index += 1
        elif operation == 1:  # Insertion to the reference (ignored in MD tag)
            query_index += length
            nc += length
        elif operation == 4:  # Soft clipping
            query_index += length
            nc += length
        elif operation == 2:  # Deletion from the reference
            if match_count > 0:
                md_tag.append(str(match_count))
                match_count = 0
            md_tag.append("^" + ref[ref_index : ref_index + length])
            ref_index += length
            nc += length

    if match_count > 0:
        md_tag.append(str(match_count))

    return "".join(md_tag), yf, zf, yc, zc, ns, nc


def filter_hits(hits, seq1, seq2):
    """
    Filter hits by the following rules:
    1. The mapping quality of the hit is greater than 0
    2. The alignment length of the hit is greater than 20
    3. The mapping length of the hit is larger than 80% of the query length
    """
    filtered_hits = []
    for hit in hits:
        q_len = len(seq1) if hit.read_num == 1 else len(seq2)
        if hit.mapq > 0 and hit.blen > 20 and hit.mlen > 0.8 * q_len:
            filtered_hits.append(hit)
    return filtered_hits


def run_mapping(name, seq1, seq2, qua1, qua2, idx0, idx_km, idx_mk, fwd_lib=True):
    # sam format specification
    # 1. QNAME: Query template NAME
    # 2. FLAG: bitwise FLAG
    # 3. RNAME: Reference sequence NAME
    # 4. POS: 1-based leftmost POSition/coordinate of clipped sequence
    # 5. MAPQ: MAPping Quality (Phred-scaled)
    # 6. CIGAR: CIGAR string
    # 7. RNEXT: Reference name of the mate/next read
    # 8. PNEXT: Position of the mate/next read
    # 9. TLEN: observed Template LENgth
    # 10. SEQ: segment SEQuence
    # 11. QUAL: ASCII of Phred-scaled base QUALity+33
    # 12. TAG: additional information

    # q_st  q_en  strand  ctg  ctg_len  r_st  r_en  mlen  blen  mapq  cg:Z:cigar_str
    if fwd_lib:
        seq1_conv = mk_conversion(seq1)
        seq2_conv = km_conversion(seq2)
    else:
        seq1_conv = km_conversion(seq1)
        seq2_conv = mk_conversion(seq2)

    mapped = []
    for idx, is_fwd in zip([idx_mk, idx_km], [True, False]):
        for hit1, hit2 in find_properly_paired_hits(
            filter_hits(
                idx.map(seq1_conv, seq2=seq2_conv, cs=False, MD=False), seq1, seq2
            ),
            fwd=is_fwd,
        ):
            tlen = max(hit1.r_en, hit2.r_en) - min(hit1.r_st, hit2.r_st)
            # flag1 and falg2 and both properly paired
            if is_fwd:
                flag1, flag2 = 99, 147
            else:
                flag1, flag2 = 83, 163
            ref1 = idx0.seq(hit1.ctg, hit1.r_st, hit1.r_en)
            ref2 = idx0.seq(hit2.ctg, hit2.r_st, hit2.r_en)
            if is_fwd:
                s1 = seq1
                s2 = mp.revcomp(seq2)
            else:
                s1 = mp.revcomp(seq1)
                s2 = seq2

            # fix soft clip
            # https://github.com/lh3/minimap2/issues/356
            cigar_str1 = hit1.cigar_str
            cigar1 = hit1.cigar
            if hit1.q_st > 0:
                cigar_str1 = f"{hit1.q_st}S" + cigar_str1
                cigar1 = [[hit1.q_st, 4]] + cigar1
            if hit1.q_en < len(s1):
                cigar_str1 = cigar_str1 + f"{len(s1) - hit1.q_en}S"
                cigar1 = cigar1 + [[len(s1) - hit1.q_en, 4]]
            cigar_str2 = hit2.cigar_str
            cigar2 = hit2.cigar
            if hit2.q_st > 0:
                cigar_str2 = f"{hit2.q_st}S" + cigar_str2
                cigar2 = [[hit2.q_st, 4]] + cigar2
            if hit2.q_en < len(s2):
                cigar_str2 = cigar_str2 + f"{len(s2) - hit2.q_en}S"
                cigar2 = cigar2 + [[len(s2) - hit2.q_en, 4]]

            md1, *debug = cal_md_and_tag(cigar1, s1, ref1, fwd_lib == is_fwd)
            md2, *_ = cal_md_and_tag(cigar2, s2, ref2, fwd_lib == is_fwd)
            # print(dict(zip(["Yf", "Zf", "Yc", "Zc", "NS", "NC"], debug)))
            # add this info into string and append ito name to debuging
            tag_info = f"__Yf:{debug[0]}+Zf:{debug[1]}+Yc:{debug[2]}+Zc:{debug[3]}+NS:{debug[4]}+NC:{debug[5]}"

            map1 = [
                name + tag_info,
                flag1,
                hit1.ctg,
                hit1.r_st + 1,
                hit1.mapq,
                cigar_str1,
                hit2.ctg,
                hit2.r_st + 1,
                tlen,
                s1,
                qua1,
                "MD:Z:" + md1,
                # strand tag
                f"ST:i:{int(is_fwd)}",
                # "cs:Z:" + hit1.cs,
            ]
            map2 = [
                name + tag_info,
                flag2,
                hit2.ctg,
                hit2.r_st + 1,
                hit2.mapq,
                cigar_str2,
                hit1.ctg,
                hit1.r_st + 1,
                -tlen,
                s2,
                qua2,
                "MD:Z:" + md2,
                f"ST:i:{int(is_fwd)}",
                # "cs:Z:" + hit2.cs,
            ]
            score = hit1.mapq + hit2.mapq
            mapped.append([score, map1, map2])

    random.shuffle(mapped)
    mapped = sorted(mapped, key=lambda x: x[0], reverse=True)
    for i, (_, map1, map2) in enumerate(mapped):
        if i > 0:
            map1[1] += 256
            map2[1] += 256
        print("\t".join(map(str, map1)))
        print("\t".join(map(str, map2)))


def map_file(ref_file, r1_file, r2_file, fwd_lib=True):
    # if mk and km file is not exist
    mk_file = ref_file + ".mk.fa"
    km_file = ref_file + ".km.fa"
    if not os.path.exists(mk_file) or not os.path.exists(km_file):
        mk_convert_file(ref_file, mk_file, km_file, include_ys_tag=False)

    # from .conversion import

    # s = a.seq("rRNA-Hsa-nucleus_locus", 100, 200)  # retrieve a subsequence from the index
    # print(mp.revcomp(s))  # reverse complement

    # Only map to the forward strand of the reference sequences. For paired-end
    # reads in the forward-reverse orientation, the first read is mapped to forward
    # strand of the reference and the second read to the reverse stand.
    # define MM_F_FOR_ONLY      (0x100000LL)
    # define MM_F_REV_ONLY      (0x200000LL)
    idx0 = mp.Aligner(
        fn_idx_in=ref_file,
        preset="sr",
        n_threads=8,
        k=10,
        w=10,
        min_cnt=0,
        min_chain_score=0,
        best_n=50,
    )
    idx_mk = mp.Aligner(
        fn_idx_in=mk_file,
        preset="sr",
        n_threads=8,
        k=10,
        w=10,
        min_cnt=0,
        min_chain_score=0,
        best_n=50,
        extra_flags=0x100000 if fwd_lib else 0x200000,
    )
    idx_km = mp.Aligner(
        fn_idx_in=km_file,
        preset="sr",
        n_threads=8,
        k=10,
        w=10,
        min_cnt=0,
        min_chain_score=0,
        best_n=50,
        extra_flags=0x200000 if fwd_lib else 0x100000,
    )
    # write sam header
    print("@HD\tVN:1.6\tSO:unsorted")
    # with qual
    for name, seq, *_ in mp.fastx_read(ref_file):
        print(f"@SQ\tSN:{name}\tLN:{len(seq)}")

    for (name1, seq1, qua1), (name2, seq2, qua2) in zip(
        mp.fastx_read(r1_file), mp.fastx_read(r2_file)
    ):
        if name1 != name2:
            raise ValueError("r1 and r2 not in the same order")
        run_mapping(name1, seq1, seq2, qua1, qua2, idx0, idx_km, idx_mk, fwd_lib)


if __name__ == "__main__":
    ref_file = "./test/ref.fa"
    r1_file = "./test/r1.fq.gz"
    r2_file = "./test/r2.fq.gz"
    map_file(ref_file, r1_file, r2_file, fwd_lib=False)

# # ## # on forward strand example
# seq_name = "test1"
# seq1 = "TTTTTTTTTTTTTTTTTTTTTTTTTTTCGGGTTGCTTGGGAATGCAGCCCAAAGCGGGTGGTAAACT"
# # seq1 = "TTGGGTTGTTTGGGGGTGTGGTGTGGGGTGGGT"
# qua1 = "IIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIII"
# # seq2 = "AGTTTACCACCCGCTTTGGGCTGCATTCCCAAGCA"
# seq2 = "AGCCCACCACCCGCCCCGGGCCGCACCCCCAAGCA"
# qua2 = "IIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIII"
#
# run_mapping(seq_name, seq1, seq2, qua1, qua2, idx0, idx_km, idx_mk, fwd_lib=True)
## # on rev strand example
## # seq1 = "GGGTCCTAACACGTGCGCTCGTGCTC"
## # seq2 = "TCTCGCCCGCCGCGCCGGGGAGGTGGAGCACGAG"



import rich_click as click

click.rich_click.COMMAND_GROUPS = {
    "coralsnake": [
        {
            "name": "Commands",
            "commands": ["prepare", "map", "liftover", "annot", "group"],
        },
    ]
}
click.rich_click.STYLE_OPTION = "bold green"
# click.rich_click.STYLE_COMMAND = "bold blue"


@click.group(
    invoke_without_command=False,
    help="Coralsnake (transcriptome mapping utils)",
    context_settings=dict(help_option_names=["-h", "--help"]),
)
@click.version_option(None, "-v", "--version")
@click.pass_context
def cli(ctx):
    pass


@cli.command(
    help="Extract primary transcript from gtf/gff file.",
    no_args_is_help=True,
    context_settings=dict(help_option_names=["-h", "--help"]),
)
@click.option("--gtf-file", "-g", "gtf_file", help="GTF file.", required=True)
@click.option("--fasta-file", "-f", "fasta_file", help="Fasta file.")
@click.option("--output-file", "-o", "output_file", help="Output file.", required=True)
@click.option("--seq-file", "-s", "seq_file", help="Sequence file.")
@click.option(
    "--sanitize",
    "-z",
    "sanitize",
    help="Sanitize name to remove specical charaters.",
    is_flag=True,
)
@click.option(
    "--with-codon",
    "-c",
    "with_codon",
    help="Include codon in the output.",
    is_flag=True,
)
@click.option(
    "--with-genename",
    "-n",
    "with_genename",
    help="Include gene name in the output.",
    is_flag=True,
)
@click.option(
    "--with-biotype",
    "-t",
    "with_biotype",
    help="Include biotype in the output.",
    is_flag=True,
)
@click.option(
    "--with-txpos",
    "-x",
    "with_txpos",
    help="Include transcript position in the output.",
    is_flag=True,
)
@click.option(
    "--filter-biotype",
    "-b",
    "filter_biotype",
    help="Filter biotype.",
    default=None,
)
@click.option(
    "--seq-upper/--seq-lower",
    "-U/-u",
    "seq_upper",
    help="Convert sequence to uppercase.",
    is_flag=True,
    default=True,
)
@click.option(
    "--line-length",
    "-l",
    "line_length",
    help="Line length. (Default: 0, no wrap)",
    default=0,
)
def prepare(
    gtf_file,
    fasta_file,
    output_file,
    seq_file,
    sanitize,
    with_codon,
    with_genename,
    with_biotype,
    with_txpos,
    filter_biotype,
    seq_upper,
    line_length,
):
    from .gtf2tx import parse_file

    if seq_file is not None and fasta_file is None:
        raise click.ClickException(
            "Error: requires --fasta-file when --seq-file is used."
        )

    parse_file(
        gtf_file,
        fasta_file,
        output_file,
        seq_file,
        sanitize,
        with_codon,
        with_genename,
        with_biotype,
        with_txpos,
        filter_biotype,
        seq_upper,
        line_length,
    )


@cli.command(
    help="Fetch genomic motif.",
    no_args_is_help=True,
    context_settings=dict(help_option_names=["-h", "--help"]),
)
@click.option("--input-bam", "-i", "input_bam", help="Input bam file.", required=True)
@click.option(
    "--output-bam", "-o", "output_bam", help="Output bam file.", required=True
)
@click.option(
    "--annotation-file", "-a", "annotation_file", help="Annotation file.", required=True
)
@click.option("--faidx-file", "-f", "faidx_file", help="Faidx file.", required=True)
@click.option("--threads", "-t", "threads", help="Threads.", default=8)
@click.option("--sort", "-s", "sort", help="Sort.", is_flag=True)
def liftover(input_bam, output_bam, annotation_file, faidx_file, threads, sort):
    from .tbam2gbam import convert_bam

    convert_bam(input_bam, output_bam, annotation_file, faidx_file, threads, sort)


@cli.command(
    help="Map reads to reference genome.",
    no_args_is_help=True,
    context_settings=dict(help_option_names=["-h", "--help"]),
)
@click.option("-r", "--ref-file", help="reference file", required=True)
@click.option("-1", "--r1-file", help="r1 file", required=True)
@click.option("-2", "--r2-file", help="r2 file", required=True)
# if --fwd is passed (or leave empty), fwd_lib will be True, otherwise if --rev is passed, fwd_lib will be False
@click.option("--fwd/--rev", "fwd_lib", default=True, help="forward or reverse library")
def map(ref_file, r1_file, r2_file, fwd_lib):
    from .mapping import map_file

    map_file(ref_file, r1_file, r2_file, fwd_lib)


@cli.command(
    help="Annotate tsv file.",
    no_args_is_help=True,
    context_settings=dict(help_option_names=["-h", "--help"]),
)
@click.option("--input-file", "-i", "input_file", help="Input file.", required=True)
@click.option("--output-file", "-o", "output_file", help="Output file.", required=True)
@click.option(
    "--annot-file", "-a", "annot_file", help="Annotation file.", required=True
)
@click.option(
    "--cols", "-c", "cols", help="Columns of Chrom,Pos,Strand", default="1,2,3"
)
@click.option("--keep-na", "-k", "keep_na", help="Keep NA.", is_flag=True)
@click.option(
    "--collapse-annot",
    "-l",
    "collapse_annot",
    help="Collapse annotation.",
    is_flag=True,
)
@click.option("--add-count", "-n", "add_count", help="Add count.", is_flag=True)
@click.option("--skip-header", "-H", "skip_header", help="Skip header.", is_flag=True)
def annot(
    input_file,
    output_file,
    annot_file,
    cols,
    keep_na,
    collapse_annot,
    add_count,
    skip_header,
):
    # --add-count is not compatible with --collapse-annot
    if add_count and collapse_annot:
        raise click.ClickException(
            "Error: --add-count is not compatible with --collapse-annot"
        )
    from .annot import run_annot

    run_annot(
        input_file,
        output_file,
        annot_file,
        cols,
        keep_na,
        collapse_annot,
        add_count,
        skip_header,
    )


@cli.command(
    help="Group and find consenus of gene.",
    no_args_is_help=True,
    context_settings=dict(help_option_names=["-h", "--help"]),
)
@click.option(
    "--fasta-file", "-f", "fasta_file", help="Fasta file.", required=True, multiple=True
)
@click.option(
    "--gtf-file",
    "-g",
    "gtf_file",
    help="GTF file.",
    required=True,
    multiple=True,
)
@click.option("--output-file", "-o", "output_file", help="Output file.", required=False)
@click.option(
    "--output-consensus",
    "-c",
    "output_consensus",
    help="Output artifical fasta file containing consenus sequences.",
    required=False,
)
@click.option(
    "--gene-name-regex", "-r", "gene_name_regex", help="Gene name regex.", default=None, type=str
)
@click.option(
    "--gene-biotype-list",
    "-b",
    "gene_biotype_list",
    help="Gene biotype list.",
    default=None,
)
@click.option(
    "--gene-length-limit",
    "-l",
    "gene_length_limit",
    help="Gene length limit.",
    default=300,
    type=int,
)
@click.option(
    "--cluster-threshold",
    "-s",
    "cluster_threshold",
    help="Clustering threshold (0-1).",
    default=0.1,
    type=float,
)
@click.option("--threads", "-t", "threads", help="Threads.", default=8)
def group(
    fasta_file,
    gtf_file,
    output_file,
    output_consensus,
    gene_name_regex,
    gene_biotype_list,
    gene_length_limit,
    cluster_threshold,
    threads,
):
    from .genegroup import group_genes

    group_genes(
        fasta_file,
        gtf_file,
        output_file,
        output_consensus,
        gene_name_regex,
        gene_biotype_list,
        gene_length_limit,
        cluster_threshold,
        threads,
    )


if __name__ == "__main__":
    cli()



from libc.stdint cimport uint8_t, int8_t
from libc.stdlib cimport free
cimport cmappy
import sys

__version__ = '2.28'

cmappy.mm_reset_timer()

cdef class Alignment:
	cdef int _ctg_len, _r_st, _r_en
	cdef int _q_st, _q_en
	cdef int _NM, _mlen, _blen
	cdef int8_t _strand, _trans_strand
	cdef uint8_t _mapq, _is_primary
	cdef int _seg_id
	cdef _ctg, _cigar, _cs, _MD # these are python objects

	def __cinit__(self, ctg, cl, cs, ce, strand, qs, qe, mapq, cigar, is_primary, mlen, blen, NM, trans_strand, seg_id, cs_str, MD_str):
		self._ctg = ctg if isinstance(ctg, str) else ctg.decode()
		self._ctg_len, self._r_st, self._r_en = cl, cs, ce
		self._strand, self._q_st, self._q_en = strand, qs, qe
		self._NM, self._mlen, self._blen = NM, mlen, blen
		self._mapq = mapq
		self._cigar = cigar
		self._is_primary = is_primary
		self._trans_strand = trans_strand
		self._seg_id = seg_id
		self._cs = cs_str
		self._MD = MD_str

	@property
	def ctg(self): return self._ctg

	@property
	def ctg_len(self): return self._ctg_len

	@property
	def r_st(self): return self._r_st

	@property
	def r_en(self): return self._r_en

	@property
	def strand(self): return self._strand

	@property
	def trans_strand(self): return self._trans_strand

	@property
	def blen(self): return self._blen

	@property
	def mlen(self): return self._mlen

	@property
	def NM(self): return self._NM

	@property
	def is_primary(self): return (self._is_primary != 0)

	@property
	def q_st(self): return self._q_st

	@property
	def q_en(self): return self._q_en

	@property
	def mapq(self): return self._mapq

	@property
	def cigar(self): return self._cigar

	@property
	def read_num(self): return self._seg_id + 1

	@property
	def cs(self): return self._cs

	@property
	def MD(self): return self._MD

	@property
	def cigar_str(self):
		return "".join(map(lambda x: str(x[0]) + 'MIDNSHP=XB'[x[1]], self._cigar))

	def __str__(self):
		if self._strand > 0: strand = '+'
		elif self._strand < 0: strand = '-'
		else: strand = '?'
		if self._is_primary != 0: tp = 'tp:A:P'
		else: tp = 'tp:A:S'
		if self._trans_strand > 0: ts = 'ts:A:+'
		elif self._trans_strand < 0: ts = 'ts:A:-'
		else: ts = 'ts:A:.'
		a = [str(self._q_st), str(self._q_en), strand, self._ctg, str(self._ctg_len), str(self._r_st), str(self._r_en),
			str(self._mlen), str(self._blen), str(self._mapq), tp, ts, "cg:Z:" + self.cigar_str]
		if self._cs != "": a.append("cs:Z:" + self._cs)
		if self._MD != "": a.append("MD:Z:" + self._MD)
		return "\t".join(a)

cdef class ThreadBuffer:
	cdef cmappy.mm_tbuf_t *_b

	def __cinit__(self):
		self._b = cmappy.mm_tbuf_init()

	def __dealloc__(self):
		cmappy.mm_tbuf_destroy(self._b)

cdef class Aligner:
	cdef cmappy.mm_idx_t *_idx
	cdef cmappy.mm_idxopt_t idx_opt
	cdef cmappy.mm_mapopt_t map_opt

	def __cinit__(self, fn_idx_in=None, preset=None, k=None, w=None, min_cnt=None, min_chain_score=None, min_dp_score=None, bw=None, bw_long=None, best_n=None, n_threads=3, fn_idx_out=None, max_frag_len=None, extra_flags=None, seq=None, scoring=None):
		self._idx = NULL
		cmappy.mm_set_opt(NULL, &self.idx_opt, &self.map_opt) # set the default options
		if preset is not None:
			cmappy.mm_set_opt(str.encode(preset), &self.idx_opt, &self.map_opt) # apply preset
		self.map_opt.flag |= 4 # always perform alignment
		self.idx_opt.batch_size = 0x7fffffffffffffffL # always build a uni-part index
		if k is not None: self.idx_opt.k = k
		if w is not None: self.idx_opt.w = w
		if min_cnt is not None: self.map_opt.min_cnt = min_cnt
		if min_chain_score is not None: self.map_opt.min_chain_score = min_chain_score
		if min_dp_score is not None: self.map_opt.min_dp_max = min_dp_score
		if bw is not None: self.map_opt.bw = bw
		if bw_long is not None: self.map_opt.bw_long = bw_long
		if best_n is not None: self.map_opt.best_n = best_n
		if max_frag_len is not None: self.map_opt.max_frag_len = max_frag_len
		if extra_flags is not None: self.map_opt.flag |= extra_flags
		if scoring is not None and len(scoring) >= 4:
			self.map_opt.a, self.map_opt.b = scoring[0], scoring[1]
			self.map_opt.q, self.map_opt.e = scoring[2], scoring[3]
			self.map_opt.q2, self.map_opt.e2 = self.map_opt.q, self.map_opt.e
			if len(scoring) >= 6:
				self.map_opt.q2, self.map_opt.e2 = scoring[4], scoring[5]
				if len(scoring) >= 7:
					self.map_opt.sc_ambi = scoring[6]

		cdef cmappy.mm_idx_reader_t *r;

		if seq is None:
			if fn_idx_out is None:
				r = cmappy.mm_idx_reader_open(str.encode(fn_idx_in), &self.idx_opt, NULL)
			else:
				r = cmappy.mm_idx_reader_open(str.encode(fn_idx_in), &self.idx_opt, str.encode(fn_idx_out))
			if r is not NULL:
				self._idx = cmappy.mm_idx_reader_read(r, n_threads) # NB: ONLY read the first part
				cmappy.mm_idx_reader_close(r)
				cmappy.mm_mapopt_update(&self.map_opt, self._idx)
				cmappy.mm_idx_index_name(self._idx)
		else:
			self._idx = cmappy.mappy_idx_seq(self.idx_opt.w, self.idx_opt.k, self.idx_opt.flag&1, self.idx_opt.bucket_bits, str.encode(seq), len(seq))
			cmappy.mm_mapopt_update(&self.map_opt, self._idx)
			self.map_opt.mid_occ = 1000 # don't filter high-occ seeds

	def __dealloc__(self):
		if self._idx is not NULL:
			cmappy.mm_idx_destroy(self._idx)

	def __bool__(self):
		return (self._idx != NULL)

	def map(self, seq, seq2=None, buf=None, cs=False, MD=False, max_frag_len=None, extra_flags=None):
		cdef cmappy.mm_reg1_t *regs
		cdef cmappy.mm_hitpy_t h
		cdef ThreadBuffer b
		cdef int n_regs
		cdef char *cs_str = NULL
		cdef int l_cs_str, m_cs_str = 0
		cdef void *km
		cdef cmappy.mm_mapopt_t map_opt

		if self._idx == NULL: return
		if ((self.map_opt.flag & 4) and (self._idx.flag & 2)): return
		map_opt = self.map_opt
		if max_frag_len is not None: map_opt.max_frag_len = max_frag_len
		if extra_flags is not None: map_opt.flag |= extra_flags

		if self._idx is NULL: return None
		if buf is None: b = ThreadBuffer()
		else: b = buf
		km = cmappy.mm_tbuf_get_km(b._b)

		_seq = seq if isinstance(seq, bytes) else seq.encode()
		if seq2 is None:
			regs = cmappy.mm_map_aux(self._idx, _seq, NULL,  &n_regs, b._b, &map_opt)
		else:
			_seq2 = seq2 if isinstance(seq2, bytes) else seq2.encode()
			regs = cmappy.mm_map_aux(self._idx, _seq, _seq2, &n_regs, b._b, &map_opt)

		try:
			i = 0
			while i < n_regs:
				cmappy.mm_reg2hitpy(self._idx, &regs[i], &h)
				cigar, _cs, _MD = [], '', ''
				for k in range(h.n_cigar32): # convert the 32-bit CIGAR encoding to Python array
					c = h.cigar32[k]
					cigar.append([c>>4, c&0xf])
				if cs or MD: # generate the cs and/or the MD tag, if requested
					_cur_seq = _seq2 if h.seg_id > 0 and seq2 is not None else _seq
					if cs:
						l_cs_str = cmappy.mm_gen_cs(km, &cs_str, &m_cs_str, self._idx, &regs[i], _cur_seq, 1)
						_cs = cs_str[:l_cs_str] if isinstance(cs_str, str) else cs_str[:l_cs_str].decode()
					if MD:
						l_cs_str = cmappy.mm_gen_MD(km, &cs_str, &m_cs_str, self._idx, &regs[i], _cur_seq)
						_MD = cs_str[:l_cs_str] if isinstance(cs_str, str) else cs_str[:l_cs_str].decode()
				yield Alignment(h.ctg, h.ctg_len, h.ctg_start, h.ctg_end, h.strand, h.qry_start, h.qry_end, h.mapq, cigar, h.is_primary, h.mlen, h.blen, h.NM, h.trans_strand, h.seg_id, _cs, _MD)
				cmappy.mm_free_reg1(&regs[i])
				i += 1
		finally:
			while i < n_regs:
				cmappy.mm_free_reg1(&regs[i])
				i += 1
			free(regs)
			free(cs_str)

	def seq(self, str name, int start=0, int end=0x7fffffff):
		cdef int l
		cdef char *s
		if self._idx == NULL: return
		if ((self.map_opt.flag & 4) and (self._idx.flag & 2)): return
		s = cmappy.mappy_fetch_seq(self._idx, name.encode(), start, end, &l)
		if l == 0: return None
		r = s[:l] if isinstance(s, str) else s[:l].decode()
		free(s)
		return r

	@property
	def k(self): return self._idx.k

	@property
	def w(self): return self._idx.w

	@property
	def n_seq(self): return self._idx.n_seq

	@property
	def seq_names(self):
		cdef char *p
		if self._idx == NULL: return
		sn = []
		for i in range(self._idx.n_seq):
			p = self._idx.seq[i].name
			s = p if isinstance(p, str) else p.decode()
			sn.append(s)
		return sn

def fastx_read(fn, read_comment=False):
	cdef cmappy.kseq_t *ks
	ks = cmappy.mm_fastx_open(str.encode(fn))
	if ks is NULL: return None
	while cmappy.kseq_read(ks) >= 0:
		if ks.qual.l > 0: qual = ks.qual.s if isinstance(ks.qual.s, str) else ks.qual.s.decode()
		else: qual = None
		name = ks.name.s if isinstance(ks.name.s, str) else ks.name.s.decode()
		seq = ks.seq.s if isinstance(ks.seq.s, str) else ks.seq.s.decode()
		if read_comment:
			if ks.comment.l > 0: comment = ks.comment.s if isinstance(ks.comment.s, str) else ks.comment.s.decode()
			else: comment = None
			yield name, seq, qual, comment
		else:
			yield name, seq, qual
	cmappy.mm_fastx_close(ks)

def revcomp(seq):
	l = len(seq)
	bseq = seq if isinstance(seq, bytes) else seq.encode()
	cdef char *s = cmappy.mappy_revcomp(l, bseq)
	r = s[:l] if isinstance(s, str) else s[:l].decode()
	free(s)
	return r

def verbose(v=None):
	if v is None: v = -1
	return cmappy.mm_verbose_level(v)
